using System;
using System.Collections.Generic;
using System.Data;
using System.Data.SqlClient;
using System.Text;
using System.Security.Principal;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;
using System.Security;
using System.Security.Permissions;
using System.Runtime.ConstrainedExecution;
using System.ComponentModel;
using Microsoft.TeamFoundation.Client;
using System.Net;

namespace TfsCmd
{
    class DbUtilities
    {
        public static SqlConnection GetConnection(DbParameters db, string database)
        {
            string connString = string.Format("server={0};database={1};", db.TFSServer, database);
            if (!string.IsNullOrEmpty(db.uid))
                connString += string.Format("uid={0};password={1}", db.uid, db.password);
            else
                connString += "trusted_connection=yes";
            return new SqlConnection(connString);
        }

        public static bool ExecuteScalar<T>(out T outParam, DbParameters db, string database, string commandText, params object[] p)
        {
            using (SqlConnection conn = GetConnection(db, database))
            {
                SqlCommand cmd = CreateSqlCommand(ref commandText, p, conn);
                conn.Open();

                try
                {
                    outParam = (T)cmd.ExecuteScalar();
                    return true;
                }
                catch (Exception)
                {
                }
            }

            outParam = default(T);
            return false;
        }

        public static void ExecuteNonQuery(DbParameters db, string database, string commandText, params object[] p)
        {
            using (SqlConnection conn = GetConnection(db, database))
            {
                SqlCommand cmd = CreateSqlCommand(ref commandText, p, conn);
                conn.Open();
                cmd.ExecuteNonQuery();
            }
        }

        public static IDataReader ExecuteDataReader(DbParameters db, string database, string commandText, params object[] p)
        {
            SqlConnection conn = GetConnection(db, database);
            SqlCommand cmd = CreateSqlCommand(ref commandText, p, conn);
            conn.Open();
            return cmd.ExecuteReader(CommandBehavior.CloseConnection);
        }

        private static SqlCommand CreateSqlCommand(ref string commandText, object[] p, SqlConnection conn)
        {
            if (p != null && p.Length > 0)
            {
                StringBuilder sb = new StringBuilder();
                string[] sqlText = commandText.Split('?');
                if (sqlText.Length != p.Length + 1)
                    throw new Exception("Invalid Sql text - missing parameters");

                for (int i = 0; i < sqlText.Length; i++)
                {
                    string text = sqlText[i];
                    if (!String.IsNullOrEmpty(text))
                    {
                        sb.Append(sqlText[i]);
                        if (i+1 < sqlText.Length)
                            sb.AppendFormat("@p{0}", i + 1);
                    }
                }

                commandText = sb.ToString();
            }

            SqlCommand cmd = new SqlCommand(commandText, conn);
            if (p != null && p.Length > 0)
            {
                for (int i = 0; i < p.Length; i++)
                {
                    cmd.Parameters.Add(string.Format("@p{0}", i + 1), SqlDbType.VarChar).Value = p[i].ToString();
                }
            }
            return cmd;
        }
    }

    class DbParameters
    {
        public string TFSServer;
        public string uid;
        public string domain;
        public string password;
        public List<string> extraParams = new List<string>();

        public DbParameters(string[] args)
        {
            for (int i = 0; i < args.Length; i++)
            {
                string arg = args[i];
                if (string.Compare(arg, "/t", true) == 0 ||
                    string.Compare(arg, "/tfs", true) == 0)
                {
                    if (args.Length > i)
                    {
                        TFSServer = args[i + 1];
                        i++;
                    }
                }
                else if (string.Compare(arg, "/u", true) == 0 ||
                    string.Compare(arg, "/user", true) == 0)
                {
                    if (args.Length > i)
                    {
                        uid = args[i + 1];
                        int idx = uid.IndexOf('\\');
                        if (idx > 0)
                        {
                            domain = uid.Substring(0, idx);
                            uid = uid.Substring(idx+1);
                        }
                        i++;
                    }
                }
                else if (string.Compare(arg, "/p", true) == 0 ||
                    string.Compare(arg, "/password", true) == 0)
                {
                    if (args.Length > i)
                    {
                        password = args[i + 1];
                        i++;
                    }
                }
                else
                    extraParams.Add(arg);
            }

            if (TFSServer == null)
            {
                TFSServer = "(local)";
            }
        }
    }

    [SecurityPermission(SecurityAction.Demand, UnmanagedCode = true)]
    class SafeUserToken : SafeHandleZeroOrMinusOneIsInvalid
    {
        [SuppressUnmanagedCodeSecurity]
        [ReliabilityContract(Consistency.WillNotCorruptState, Cer.Success)]
        [DllImport("kernel32")]
        private static extern bool CloseHandle(IntPtr handle);

        public SafeUserToken () : base(true)
	    {
	    }

        override protected bool ReleaseHandle()
        {
            return CloseHandle(handle);
        }

    }

    public enum LogonType : int
    {
        /// <summary>
        /// This logon type is intended for users who will be interactively using the computer, such as a user being logged on  
        /// by a terminal server, remote shell, or similar process.
        /// This logon type has the additional expense of caching logon information for disconnected operations;
        /// therefore, it is inappropriate for some client/server applications,
        /// such as a mail server.
        /// </summary>
        LOGON32_LOGON_INTERACTIVE = 2,

        /// <summary>
        /// This logon type is intended for high performance servers to authenticate plaintext passwords.

        /// The LogonUser function does not cache credentials for this logon type.
        /// </summary>
        LOGON32_LOGON_NETWORK = 3,

        /// <summary>
        /// This logon type is intended for batch servers, where processes may be executing on behalf of a user without
        /// their direct intervention. This type is also for higher performance servers that process many plaintext
        /// authentication attempts at a time, such as mail or Web servers.
        /// The LogonUser function does not cache credentials for this logon type.
        /// </summary>
        LOGON32_LOGON_BATCH = 4,

        /// <summary>
        /// Indicates a service-type logon. The account provided must have the service privilege enabled.
        /// </summary>
        LOGON32_LOGON_SERVICE = 5,

        /// <summary>
        /// This logon type is for GINA DLLs that log on users who will be interactively using the computer.
        /// This logon type can generate a unique audit record that shows when the workstation was unlocked.
        /// </summary>
        LOGON32_LOGON_UNLOCK = 7,

        /// <summary>
        /// This logon type preserves the name and password in the authentication package, which allows the server to make
        /// connections to other network servers while impersonating the client. A server can accept plaintext credentials
        /// from a client, call LogonUser, verify that the user can access the system across the network, and still
        /// communicate with other servers.
        /// NOTE: Windows NT:  This value is not supported.
        /// </summary>
        LOGON32_LOGON_NETWORK_CLEARTEXT = 8,

        /// <summary>
        /// This logon type allows the caller to clone its current token and specify new credentials for outbound connections.
        /// The new logon session has the same local identifier but uses different credentials for other network connections.
        /// NOTE: This logon type is supported only by the LOGON32_PROVIDER_WINNT50 logon provider.
        /// NOTE: Windows NT:  This value is not supported.
        /// </summary>
        LOGON32_LOGON_NEW_CREDENTIALS = 9,
    }

    public enum LogonProvider : int
    {
        /// <summary>
        /// Use the standard logon provider for the system.
        /// The default security provider is negotiate, unless you pass NULL for the domain name and the user name
        /// is not in UPN format. In this case, the default provider is NTLM.
        /// NOTE: Windows 2000/NT:   The default security provider is NTLM.
        /// </summary>
        LOGON32_PROVIDER_DEFAULT = 0,
    }

    [SecurityPermission(SecurityAction.Demand, UnmanagedCode=true)]
    class ImpersonateUser : IDisposable
    {
        private WindowsImpersonationContext _context = null;
        private SafeUserToken _secToken = null;

        [DllImport("advapi32.dll", SetLastError=true)]
        public static extern bool LogonUser(string lpszUsername, string lpszDomain, string lpszPassword, LogonType dwLogonType, LogonProvider dwLogonProvider, out SafeUserToken phToken);

        public ImpersonateUser(string userName, string domain, string password)
            : this(userName, domain, password, LogonType.LOGON32_LOGON_INTERACTIVE, LogonProvider.LOGON32_PROVIDER_DEFAULT)
        {
        }

        public ImpersonateUser(string userName, string domain, string password, LogonType lType, LogonProvider lProvider)
        {
            // Validate input
            if (String.IsNullOrEmpty(userName))
                throw new ArgumentNullException("userName");
            if (String.IsNullOrEmpty(password))
                throw new ArgumentNullException("password");

            // Get a user token
            if (!LogonUser(userName, domain, password, lType, lProvider, out _secToken))
                throw new Win32Exception(Marshal.GetLastWin32Error());

            // Impersonate this user.
            _context = WindowsIdentity.Impersonate(_secToken.DangerousGetHandle());
        }

        public void UndoImpersonation()
        {
            if (_context != null)
            {
                _context.Undo();
                _context = null;
            }
            if (_secToken != null)
            {
                _secToken.Dispose();
                _secToken = null;
            }
        }

        public void Dispose()
        {
            UndoImpersonation();
        }
    }
}
